#!/usr/bin/env python3
#-*- coding:utf-8 -*-
##
## main.py
##
##  Created on: Fev 20, 2020
##      Author: Antonio Morgado
##      E-mail: ajrmorgado@gmail.com
##
# Copyright (C) 2020, Antonio Morgado, Mikolas Janota
#
# This file is part of dtfinder.
#
# dtfinder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dtfinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dtfinder.  If not, see <https://www.gnu.org/licenses/>.
#
#==============================================================================
#
# required imports

import sys
import getopt
import os
import inspect
import math
import time
from pysat.solvers import Solver
from pysat.formula import IDPool , CNF, WCNF
from pysat.card import *
from pysat.examples.rc2 import RC2

import mytrie

sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../scripts/'))

import enum_topologies

import functools
print = functools.partial(print, flush=True)

sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../greedy/'))

import run_sklearn

#
#==============================================================================
# start the encoding
class DT1(CNF, object):
    #
    #--------------------------------------------------------------------------
    def create_variables(self, topv):
        self.vpool = IDPool(start_from=topv+1)
        
        self.vvar = lambda i: self.vpool.id('v{0}'.format(i))
        self.lvar = lambda i,j: self.vpool.id('l{0} {1}'.format(i,j))
        self.rvar = lambda i,j: self.vpool.id('r{0} {1}'.format(i,j))
        self.pvar = lambda j,i: self.vpool.id('p{0} {1}'.format(j,i))
        
        self.avar = lambda i,j: self.vpool.id('a{0} {1}'.format(i,j))
        self.uvar = lambda i,j: self.vpool.id('u{0} {1}'.format(i,j))
        self.d0var = lambda i,j: self.vpool.id('d0 {0} {1}'.format(i,j))
        self.d1var = lambda i,j: self.vpool.id('d1 {0} {1}'.format(i,j))
        self.cvar = lambda i: self.vpool.id('c{0}'.format(i))
        
        self.ubvar = lambda i,j: self.vpool.id('ub {0} {1}'.format(i,j))
        self.lbvar = lambda i,j: self.vpool.id('lb {0} {1}'.format(i,j))
        
        self.matchvar = lambda s,i: self.vpool.id('match{0} {1}'.format(s,i))
        
        # aux vars
        
        self.pd0var = lambda i,j,r: self.vpool.id('pd0 {0} {1} {2}'.format(i,j,r))
        self.arvar = lambda i,j,r: self.vpool.id('ar{0} {1} {2}'.format(i,j,r))
        self.pd1var = lambda i,j,r: self.vpool.id('pd1 {0} {1} {2}'.format(i,j,r))
        self.alvar = lambda i,j,r: self.vpool.id('al{0} {1} {2}'.format(i,j,r))
        
        self.upvar = lambda i,j,r: self.vpool.id('up{0} {1} {2}'.format(i,j,r))
    #
    #--------------------------------------------------------------------------
    def print_variables(self):
        print('M={0}'.format(self.M))
        print('K={0}'.format(self.K))
        print('N={0}'.format(self.N))
        
        print('# -- binary tree variables -- #')
        
        for i in range(1,self.N+1):
            print('v({0})={1}'.format(i,self.vvar(i)))
            
        for i in range(1,self.N+1):
            for j in range(i+1, min(2*i, self.N-1)+1):
                if j % 2 == 0:
                    print('l({0},{1})={2}'.format(i,j,self.lvar(i,j)))
                    
        for i in range(1,self.N+1):
            for j in range(i+2, min(2*i+1, self.N)+1):
                if j % 2 != 0:
                    print('r({0},{1})={2}'.format(i,j,self.rvar(i,j)))
                    
        for j in range(2,self.N+1):
            for i in range(math.floor(j/2),min(j-1, self.N)+1):
                if (i + 1  != j) or i % 2 != 0:
                    print('p({0},{1})={2}'.format(j,i,self.pvar(j,i)))
        
        print('# -- decision tree variables -- #')
        
        for r in range(1, self.K+1):
            for j in range(1, self.N+1):
                print('a({0},{1})={2}'.format(r,j,self.avar(r,j)))
        
        for r in range(1, self.K+1):
            for j in range(1, self.N+1):
                print('u({0},{1})={2}'.format(r,j,self.uvar(r,j)))
                
        for r in range(1, self.K+1):
            for j in range(1, self.N+1):
                print('d0({0},{1})={2}'.format(r,j,self.d0var(r,j)))
                
        for r in range(1, self.K+1):
            for j in range(1, self.N+1):
                print('d1({0},{1})={2}'.format(r,j,self.d1var(r,j)))
                
        for j in range(1, self.N+1):
            print('c({0})={1}'.format(j, self.cvar(j)))
            
        print('# -- auxiliary variables -- #')
            
        for r in range(1, self.K+1):
            for j in range(2, self.N+1):
                for i in range(math.floor(j/2), j):
                    if (i + 1  != j) or i % 2 != 0:
                        print('pd0({0},{1},{2})={3}'.format(i,j,r,self.pd0var(i,j,r)))
                        if j % 2 != 0:
                            print('ar({0},{1},{2})={3}'.format(i,j,r,self.arvar(i,j,r)))
                        
        for r in range(1, self.K+1):
            for j in range(2, self.N+1):
                for i in range(math.floor(j/2), j):
                    if (i + 1  != j) or i % 2 != 0:
                        print('pd1({0},{1},{2})={3}'.format(i,j,r,self.pd1var(i,j,r)))
                        if j % 2 == 0:
                            print('al({0},{1},{2})={3}'.format(i,j,r,self.alvar(i,j,r)))
        
        for r in range(1, self.K+1):
            for j in range(2, self.N+1):
                for i in range(math.floor(j/2), j):
                    if (i + 1  != j) or i % 2 != 0:
                        print('up({0},{1},{2})={3}'.format(i,j,r,self.upvar(i,j,r)))
        
        if self.ubencd > 0:
            print('# -- ub encoding variables -- #')
            for i in range(1, self.N+1):
                for t in range(0, math.floor(i/2)):
                    print('ub({1},{0})={2}'.format(i,t,self.ubvar(t,i)))
                        
        if self.lbencd > 0:
            print('# -- lb encoding variables -- #')
            for i in range(1, self.N+1):
                for t in range(0, i):
                    print('lb({1},{0})={2}'.format(i,t,self.lbvar(t,i)))
    #
    #--------------------------------------------------------------------------
    def __init__(self, examples, satsolver, alg, featnames, cencd=0, nnodes=3, ubenc=0, lbenc=0, quasipure=None, matchencd = 0, trie = None, SKL=None, SKL_TOPO=None, tpldepth=3, ordertopologies=0, useassumps=500, verb=0, topv=0):
        super(DT1,self).__init__()
        
        self.samples = examples
        self.algorithm = alg
        self.card = cencd
        self.verbosity = verb
        self.quasipure = quasipure
        self.M = len(examples)
        self.K = len(examples[0])-1
        self.N = nnodes
        self.solver = satsolver
        self.ubencd = ubenc
        self.lbencd = lbenc
        self.matchencd = matchencd
        self.trie = trie
        self.SKL = SKL
        self.SKL_TOPO = SKL_TOPO
        self.tpldepth = tpldepth
        self.ordertopologies = ordertopologies
        self.useassumps = useassumps
        
        #if self.N == -1:
        #    self.N = 2*len(examples) - 1
        #if self.N % 2 == 0:
        #    self.N = self.N-1
        
        
        self.ub = -1
        self.model = []
        self.fnames = featnames
        
        
        self.create_variables(topv)
        #self.print_variables()
    #
    #--------------------------------------------------------------------------
    def atMostOne(self, literals):
    
        encType = EncType.pairwise
        
        if len(literals) > 7:
        
            if self.card == 1:
                encType = EncType.seqcounter
            elif self.card == 2:
                encType = EncType.sortnetwrk
            elif self.card == 3:
                encType = EncType.cardnetwrk
            elif self.card == 4:
                encType = EncType.bitwise
            elif self.card == 5:
                encType = EncType.ladder
            elif self.card == 6:
                encType = EncType.totalizer
            elif self.card == 7:
                encType = EncType.mtotalizer
            elif self.card == 8:
                encType = EncType.kmtotalizer
            
        am1 = CardEnc.atmost(lits=literals, top_id=self.vpool._next(), encoding=encType)
        self.vpool.occupy(self.vpool._next(), am1.nv)
        return am1
    #
    #--------------------------------------------------------------------------
    def encode_tree(self, cnf):
        # v(1) is false
        cnf.append([-self.vvar(1)])
        
        #print('#####\n v(1) is false \n', cnf.clauses)
        
        # a leaf has no children
        for i in range(2,self.N+1):
            for j in range(i+1, min(2*i, self.N-1)+1):
                if j % 2 ==0:
                    cnf.append([-self.vvar(i), -self.lvar(i,j)])
                    
        #print('#####\n a leaf has no children \n', cnf.clauses)
        
        # left and right child have consecutive numbers
        for i in range(1,self.N+1):
            for j in range(i+1, min(2*i, self.N-1)+1):
                if j % 2 ==0:
                    cnf.append([-self.lvar(i,j), self.rvar(i,j+1)])
                    cnf.append([self.lvar(i,j), -self.rvar(i,j+1)])
                    
        #print('#####\n left and right child have consecutive numbers \n', cnf.clauses)
                    
        # non-leaf node has a child
        for i in range(1,self.N+1):
            lits = []
            for j in range(i+1, min(2*i, self.N-1)+1):
                if j % 2 ==0:
                    lits.append(self.lvar(i,j))
            vi = self.vvar(i)
            am1=self.atMostOne(lits)
            for x in range(len(am1.clauses)):
                am1.clauses[x].append(vi)
                cnf.append(am1.clauses[x])
            lits.append(vi)
            cnf.append(lits)
            
        #print('#####\n non-leaf node has a child \n', cnf.clauses)
            
        # a parent must have a child
        for i in range(1,self.N+1):
            for j in range(i+1, min(2*i, self.N-1)+1):
                if j % 2 == 0:
                    pji = self.pvar(j,i)
                    lij = self.lvar(i,j)
                    cnf.append([-pji,lij])
                    cnf.append([pji,-lij])
        for i in range(1,self.N+1):
            for j in range(i+2, min(2*i+1, self.N)+1):
                if j % 2 != 0:
                    pji = self.pvar(j,i)
                    rij = self.rvar(i,j)
                    cnf.append([-pji,rij])
                    cnf.append([pji,-rij])
                    
        #print('#####\n a parent must have a child \n', cnf.clauses)
                    
        # nodes must have a parent (except first)
        for j in range(2, self.N+1):
            lits = []
            for i in range(math.floor(j/2),min(j-1, self.N)+1):
                if (i+1) != j or i % 2 != 0:
                    #print(i,j,self.pvar(j,i))
                    lits.append(self.pvar(j,i))
            #print(lits)
            cnf.append(lits)
            cnf.extend(self.atMostOne(lits).clauses)
            
        #print('#####\n nodes must have a parent (except first) \n', cnf.clauses)
    #
    #--------------------------------------------------------------------------
    def encode_features(self, cnf):
        # discriminate a feature for value 0 at node j
        for r in range(1,self.K+1):
            # j=1
            cnf.append([-self.d0var(r,1)])
            # j=2..N
            for j in range(2, self.N+1):
                #gather the literals of the constraint
                lits = []
                for i in range(math.floor(j/2), j):
                    if (i + 1  != j) or i % 2 != 0:
                        # define the constraints of pd0var(i,j,r)
                        cnf.append([-self.pd0var(i,j,r), self.pvar(j,i)])
                        cnf.append([-self.pd0var(i,j,r), self.d0var(r,i)])
                        cnf.append([self.pd0var(i,j,r), -self.pvar(j,i), -self.d0var(r,i)])
                        # add the literal pd0var(i,j,r)
                        lits.append(self.pd0var(i,j,r))
                        if j % 2 != 0:
                            # define the constraints of arvar(i,j,r)
                            cnf.append([-self.arvar(i,j,r), self.avar(r,i)])
                            cnf.append([-self.arvar(i,j,r), self.rvar(i,j)])
                            cnf.append([self.arvar(i,j,r), -self.avar(r,i), -self.rvar(i,j)])
                            # add the literal arvar(i,j,r)
                            lits.append(self.arvar(i,j,r))
                # create the constraint based on the literals gathered
                for l in lits:
                    cnf.append([self.d0var(r,j), -l])
                lits.append(-self.d0var(r,j))
                cnf.append(lits)
        
        #print('#####\n discriminate a feature for value 0 at node j \n', cnf.clauses)
        
        # discriminate a feature for value 1 at node j
        for r in range(1,self.K+1):
            # j=1
            cnf.append([-self.d1var(r,1)])
            # j=2..N
            for j in range(2, self.N+1):
                #gather the literals of the constraint
                lits = []
                for i in range(math.floor(j/2), j):
                    if (i + 1  != j) or i % 2 != 0:
                        # define the constraints of pd1var(i,j,r)
                        cnf.append([-self.pd1var(i,j,r), self.pvar(j,i)])
                        cnf.append([-self.pd1var(i,j,r), self.d1var(r,i)])
                        cnf.append([self.pd1var(i,j,r), -self.pvar(j,i), -self.d1var(r,i)])
                        # add the literal pd1var(i,j,r)
                        lits.append(self.pd1var(i,j,r))
                        if j % 2 == 0:
                            # define the constraints of alvar(i,j,r)
                            cnf.append([-self.alvar(i,j,r), self.avar(r,i)])
                            cnf.append([-self.alvar(i,j,r), self.lvar(i,j)])
                            cnf.append([self.alvar(i,j,r), -self.avar(r,i), -self.lvar(i,j)])
                            # add the literal arvar(i,j,r)
                            lits.append(self.alvar(i,j,r))
                # create the constraint based on the literals gathered
                for l in lits:
                    cnf.append([self.d1var(r,j), -l])
                lits.append(-self.d1var(r,j))
                cnf.append(lits)
        
        #print('#####\n discriminate a feature for value 1 at node j \n', cnf.clauses)
        
        # using feature r at node j
        ## disallow feature if previously used
        for r in range(1, self.K+1):
            for j in range(1, self.N+1):
                for i in range(math.floor(j/2), j):
                    if (i+1) != j or i % 2 != 0:
                        cnf.append([-self.uvar(r,i), -self.pvar(j,i), -self.avar(r,j)])
        
        #print('#####\n disallow feature if previously used \n', cnf.clauses )
        
        ## define variable u
        for r in range(1, self.K+1):
            for j in range(1, self.N+1):
                # gather the literals of the constraint
                lits = [self.avar(r,j)]
                for i in range(math.floor(j/2), j):
                    if (i+1) != j or i % 2 != 0:
                        # define the constraints of upvar(i,j,r)
                        cnf.append([-self.upvar(i,j,r), self.uvar(r,i)])
                        cnf.append([-self.upvar(i,j,r), self.pvar(j,i)])
                        cnf.append([self.upvar(i,j,r), -self.uvar(r,i), -self.pvar(j,i)])
                        #add the literal upvar(i,j,r)
                        lits.append(self.upvar(i,j,r))
                # create the constraint based on the literals gathered
                for l in lits:
                    cnf.append([self.uvar(r,j), -l])
                lits.append(-self.uvar(r,j))
                cnf.append(lits)
        
        #print('#####\n define variable u\n', cnf.clauses)
        
        # non-leaf node have exactly one feature assigned
        for j in range(1, self.N+1):
            lits = []
            for r in range(1, self.K+1):
                lits.append(self.avar(r,j))
            am1 = self.atMostOne(lits)
            for cl in am1.clauses:
                cl.append(self.vvar(j))
                cnf.append(cl)
            lits.append(self.vvar(j))
            cnf.append(lits)
        
        #print('#####\n non-leaf node have exactly one feature assigned \n', cnf.clauses)
        
        # leaf nodes are not assigned features
        for j in range(1, self.N+1):
            for r in range(1, self.K+1):
                cnf.append([-self.vvar(j), -self.avar(r,j)])
                
        #print('#####\n leaf nodes are not assigned features \n', cnf.clauses,'\n\n')
    #
    #--------------------------------------------------------------------------
    # encode the samples
    def encode_samples(self, cnf):
    
        if self.matchencd == 3:
            self.samples = None
            
            for j in range(1, self.N+1):
                stack = [self.trie.root]
                m0j = self.matchvar(self.trie.root.nid, j)
                cnf.append([m0j])
                while len(stack) > 0:
                    node = stack.pop()
                    
                    if node.child0 is not None:
                    
                        feats = [node.feat]
                        dir = [0]
                        nid = node.child0.nid
                        nstack = [node.child0]

                        while len(nstack) > 0:
                            nnode = nstack.pop()
                            if nnode.child0 is None and nnode.child1 is None:
                                nid =  nnode.nid
                                stack.append(nnode)
                            elif nnode.child0 is not None and nnode.child1 is not None:
                                nid =  nnode.nid
                                stack.append(nnode)
                            elif nnode.child0 is not None:
                                feats.append(nnode.feat)
                                dir.append(0)
                                nstack.append(nnode.child0)
                            else:
                                feats.append(nnode.feat)
                                dir.append(1)
                                nstack.append(nnode.child1)

                        mpj = self.matchvar(node.nid, j)
                        msj = self.matchvar(nid, j)
                            
                        #cnf.append([-msj, mpj])
                        cl = [msj, -mpj]
                        for i in range(len(feats)):
                            f = feats[i]
                            d = dir[i]
                            r = f + 1
                            if d == 0:
                                d = self.d0var(r,j)
                            else:
                                d = self.d1var(r,j)
                            #cnf.append([-msj, -d])
                            cl.append(d)
                        cnf.append(cl)
                        
                    if node.child1 is not None:
                        
                        feats = [node.feat]
                        dir = [1]
                        nid = node.child1.nid
                        nstack = [node.child1]

                        while len(nstack) > 0:
                            nnode = nstack.pop()
                            if nnode.child0 is None and nnode.child1 is None:
                                nid =  nnode.nid
                                stack.append(nnode)
                            elif nnode.child0 is not None and nnode.child1 is not None:
                                nid =  nnode.nid
                                stack.append(nnode)
                            elif nnode.child0 is not None:
                                feats.append(nnode.feat)
                                dir.append(0)
                                nstack.append(nnode.child0)
                            else:
                                feats.append(nnode.feat)
                                dir.append(1)
                                nstack.append(nnode.child1)

                        mpj = self.matchvar(node.nid, j)
                        msj = self.matchvar(nid, j)
                            
                        #cnf.append([-msj, mpj])
                        cl = [msj, -mpj]
                        for i in range(len(feats)):
                            f = feats[i]
                            d = dir[i]
                            r = f + 1
                            if d == 0:
                                d = self.d0var(r,j)
                            else:
                                d = self.d1var(r,j)
                            #cnf.append([-msj, -d])
                            cl.append(d)
                        cnf.append(cl)
                    
                    if node.child0 is None and node.child1 is None:
                    
                        msj = self.matchvar(node.nid, j)
                        
                        clt = [-msj, -self.vvar(j)]
                        if node.c == 1:
                            clt.append(self.cvar(j))
                        else:
                            clt.append(-self.cvar(j))
                        cnf.append(clt)
                        
        elif self.matchencd == 2:
            self.samples = None
            
            for j in range(1, self.N+1):
                stack = [self.trie.root]
                m0j = self.matchvar(self.trie.root.nid, j)
                cnf.append([m0j])
                
                while len(stack) > 0:
                    node = stack.pop()
                    if node.child0 is not None:
                        stack.append(node.child0)
                        
                        mpj = self.matchvar(node.nid, j)
                        msj = self.matchvar(node.child0.nid, j)
                        r = node.feat + 1
                        d = self.d0var(r,j)
                        
                        #cnf.append([-msj, mpj])
                        #cnf.append([-msj, -d])
                        cnf.append([msj, -mpj, d])
                        
                    if node.child1 is not None:
                        stack.append(node.child1)
                        
                        mpj = self.matchvar(node.nid, j)
                        msj = self.matchvar(node.child1.nid, j)
                        r = node.feat + 1
                        d = self.d1var(r,j)
                        
                        #cnf.append([-msj, mpj])
                        #cnf.append([-msj, -d])
                        cnf.append([msj, -mpj, d])
                        
                    if node.child0 is None and node.child1 is None:
                    
                        msj = self.matchvar(node.nid, j)
                        
                        clt = [-msj, -self.vvar(j)]
                        if node.c == 1:
                            clt.append(self.cvar(j))
                        else:
                            clt.append(-self.cvar(j))
                        cnf.append(clt)
                        
            #print('#####\n samples \n', cnf.clauses,'\n\n')
        elif self.matchencd == 1:
            for j in range(1, self.N+1):
                for s in range(len(self.samples)):
                    sample = self.samples[s]
                    msj = self.matchvar(s,j)
                    cl = [msj]
                    for r in range(1, self.K+1):
                        if sample[r-1] == 0:
                            cl.append(self.d0var(r,j))
                            #cnf.append([-msj, -self.d0var(r,j)])
                        else:
                            cl.append(self.d1var(r,j))
                            #cnf.append([-msj, -self.d1var(r,j)])
                    cnf.append(cl)
                    clt = [-msj, -self.vvar(j)]
                    if sample[self.K] == 1:
                        clt.append(self.cvar(j))
                    else:
                        clt.append(-self.cvar(j))
                    cnf.append(clt)
            
            #print('#####\n samples \n', cnf.clauses,'\n\n')
        else:
            for sample in self.samples:
                for j in range(1, self.N+1):
                    cl = [-self.vvar(j)]
                    if sample[self.K] == 1:
                        cl.append(self.cvar(j))
                    else:
                        cl.append(-self.cvar(j))
                    for r in range(1, self.K+1):
                        if sample[r-1] == 0:
                            cl.append(self.d0var(r,j))
                        else:
                            cl.append(self.d1var(r,j))
                    cnf.append(cl)
            #print('#####\n samples \n', cnf.clauses,'\n\n')
    #
    #--------------------------------------------------------------------------
    #
    def encode_ub(self, cnf):
        for i in range(1, self.N+1):
            cnf.append([self.ubvar(0,i)])
            for t in range(1, math.floor(i/2)):
                aux = self.vpool.id('ubv {} {}'.format(t-1,i))
                cnf.append([-aux, self.ubvar(t-1,i-1)])
                cnf.append([-aux, self.vvar(i)])
                cnf.append([aux, -self.ubvar(t-1,i-1), -self.vvar(i)])
                
                cnf.append([-self.ubvar(t,i), self.ubvar(t,i-1), aux])
                cnf.append([self.ubvar(t,i), -self.ubvar(t,i-1)])
                cnf.append([self.ubvar(t,i), -aux])
                
        for i in range(1, self.N+1):
            for t in range(1, math.floor(i/2)):
                cnf.append([-self.ubvar(t,i), -self.lvar(i, 2*(i-t+1))])
                cnf.append([-self.ubvar(t,i), -self.rvar(i, 2*(i-t+1)+1)])
    #
    #--------------------------------------------------------------------------
    #
    def encode_lb(self, cnf):
        for i in range(1, self.N+1):
            cnf.append([self.lbvar(0,i)])
            for t in range(1, i):
                aux = self.vpool.id('lbv {} {}'.format(t-1,i))
                cnf.append([-aux, self.lbvar(t-1,i-1)])
                cnf.append([-aux, -self.vvar(i)])
                cnf.append([aux, -self.lbvar(t-1,i-1), self.vvar(i)])
                
                cnf.append([-self.lbvar(t,i), self.lbvar(t,i-1), aux])
                cnf.append([self.lbvar(t,i), -self.lbvar(t,i-1)])
                cnf.append([self.lbvar(t,i), -aux])
                
        for i in range(1, self.N+1):
            for t in range(math.ceil(i/2)+1,i+1):
                cnf.append([-self.lbvar(t,i), -self.lvar(i, 2*(t-1))])
                cnf.append([-self.lbvar(t,i), -self.lvar(i, 2*t-1)])
    #
    #--------------------------------------------------------------------------
    #
    def encode_quasipure(self, cnf):
    
        return None
        
        for qpindex in self.quasipure.keys():
            k = qpindex+1
            fcount = self.quasipure[qpindex]
            for i in range(1, self.N+1):
                if fcount[0][0] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), self.xvar(i,self.maxpath()), self.cvar(i)])
                if fcount[0][1] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), -self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), self.xvar(i,self.maxpath()), -self.cvar(i)])
                if fcount[1][0] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), -self.xvar(i,self.maxpath()), self.cvar(i)])
                if fcount[1][1] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), -self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), -self.xvar(i,self.maxpath()), -self.cvar(i)])
    #
    #--------------------------------------------------------------------------
    # start the encoding in a cnf file
    def encode(self):
        cnf = CNF()
        self.encode_tree(cnf)
        self.encode_features(cnf)
        self.encode_samples(cnf)
        if self.ubencd > 0:
            self.encode_ub(cnf)
        if self.lbencd > 0:
            self.encode_lb(cnf)
        if self.quasipure is not None:
            self.encode_quasipure(cnf)
        if self.verbosity:
            print('# n.clauses: {}'.format(len(cnf.clauses)))
        return cnf
    #
    #--------------------------------------------------------------------------
    def save_model(self, model, n):
        self.ub = n
        self.model = model
    #
    #--------------------------------------------------------------------------
    def sklearn_tree_to_sol(self, clf):
        n_nodes = clf.tree_.node_count
        children_left = clf.tree_.children_left
        children_right = clf.tree_.children_right
        feature = clf.tree_.feature
        threshold = clf.tree_.threshold
        value = clf.tree_.value
        rv = []
        for node_id in range(n_nodes):
            if (children_left[node_id] != children_right[node_id]):
                rv.append("a {} {}".format(self.fnames[feature[node_id]], node_id+1))
                rv.append("l {} {}".format(node_id+1, children_left[node_id]+1))
                rv.append("r {} {}".format(node_id+1, children_right[node_id]+1))
            else:
                rv.append("c {} {}".format(node_id+1, 0 if value[node_id][0].item(0) else 1))
        return rv
    #
    #--------------------------------------------------------------------------
    def report_solution(self):
        
        if self.ub == -1:
            print("UNSAT")
            return -1
        #print('# minimum number of nodes:', self.ub, self.model)
        
        print('# minimum n paths:', int((self.ub+1)/2), 'minimum n nodes:', self.ub)
        
        if self.model is None:
            SKL_DEC_TREE = self.sklearn_tree_to_sol(self.SKL)
            for l in SKL_DEC_TREE:
                print(l)
            return 1
            
        if self.model == []:
            return None
        
        for i in range(1,self.ub+1):
            if self.model[self.vvar(i)-1] > 0:
                sgn = 0
                if self.model[self.cvar(i)-1] > 0:
                    sgn = 1
                print('c {} {}'.format(i,sgn))
                
        for r in range(1, self.K+1):
            for j in range(1, self.ub+1):
                if self.model[self.avar(r,j)-1] > 0:
                    print('a {} {}'.format(self.fnames[r-1],j))
                    
        for i in range(1,self.ub+1):
            for j in range(i+1, min(2*i, self.ub-1)+1):
                if j % 2 == 0:
                    if self.model[self.lvar(i,j)-1] > 0:
                        print('l {} {}'.format(i,j))
                    
        for i in range(1,self.ub+1):
            for j in range(i+2, min(2*i+1, self.ub)+1):
                if j % 2 != 0:
                    if self.model[self.rvar(i,j)-1] > 0:
                        print('r {} {}'.format(i,j))
        return 1
    #
    #--------------------------------------------------------------------------
    def solve_lssu(self):
        ubinit = self.N
        
        while self.N > 1 :
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
                
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                self.N = self.N-2
            else:
                cnf = self.encode()
                oracle = Solver(name = self.solver, bootstrap_with = cnf.clauses)
                t = time.time()
                res = oracle.solve()
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    return self.report_solution()
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    self.N = self.N-2
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def solve_lsus(self):
        stop = False
        ubinit = self.N
        self.N = 3
        while not stop :
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                stop = True
            else:
                cnf = self.encode()
                oracle = Solver(name = self.solver, bootstrap_with = cnf.clauses)
                t = time.time()
                res = oracle.solve()
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    self.N = self.N+2
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    stop = True
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def solve_bs(self):
        stop = False
        ubinit = self.N
        ub = self.N+2
        lb = 1
        while not stop :
            self.N = int((ub + lb)/2)
            if self.N % 2 == 0:
                self.N = self.N-1
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
            else:
                cnf = self.encode()
                oracle = Solver(name = self.solver, bootstrap_with = cnf.clauses)
                t = time.time()
                res = oracle.solve()
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
            if ub == lb + 2:
                stop = True
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def solve_prgsu(self):
        stop = False
        ubinit = self.N
        ub = self.N+2
        lb = 1
        p = 2
        i = 1
        while not stop :
            self.N = max(ub-pow(2,i), lb+2)
            if self.N % 2 == 0:
                self.N = self.N-1
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
            if self.N == ubinit:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
                i = i+1
            else:
                cnf = self.encode()
                oracle = Solver(name = self.solver, bootstrap_with = cnf.clauses)
                t = time.time()
                res = oracle.solve()
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                    i = 1
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
                    i = i+1
            if ub == lb + 2:
                stop = True
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def solve_prgus(self):
        stop = False
        ubinit = self.N
        ub = self.N+2
        lb = 1
        p = 2
        i = 1
        while not stop :
            self.N = min(lb+pow(2,i), ub-2)
            if self.N % 2 == 0:
                self.N = self.N-1
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
            if self.N == ubinit:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
                i = 1
            else:
                cnf = self.encode()
                oracle = Solver(name = self.solver, bootstrap_with = cnf.clauses)
                t = time.time()
                res = oracle.solve()
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                    i = i+1
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
                    i = 1
            if ub == lb + 2:
                stop = True
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def satinit(self, cnf, useassumps):
        if useassumps:
            if self.solver == 'g3' or self.solver == 'g4':
                return Solver(name = self.solver, bootstrap_with = cnf.clauses, incr = True)
        return Solver(name = self.solver, bootstrap_with = cnf.clauses)
    #
    #--------------------------------------------------------------------------
    def settopology(self, root):
    
        assumps = []
        stack = [root]
        while len(stack) > 0:
            node = stack.pop()
                
            if node.card == 1:
                assumps.append(self.vvar(node.node_id))
            elif node.node_id > 1:
                assumps.append(-self.vvar(node.node_id))
            
            if node.left is not None:
                stack.append(node.left)
                assumps.append(self.lvar(node.node_id,node.left.node_id))
                assumps.append(self.pvar(node.left.node_id,node.node_id))
            
            if node.right is not None:
                stack.append(node.right)
                assumps.append(self.rvar(node.node_id,node.right.node_id))
                assumps.append(self.pvar(node.right.node_id,node.node_id))
            
        return assumps
    #
    #--------------------------------------------------------------------------
    def et_satsolve(self, cnf):
    
        useassumps = self.useassumps
        
        oracle = None
        res = None
        
        tpldepth = self.tpldepth
        
        if self.N <= 15:
            tpldepth = self.N
        elif self.N == 17:
            tpldepth = 4
        elif self.N <= 21:
            tpldepth = 3
          
        tgen = time.time()
        topologies = enum_topologies.enum1(self.N, tpldepth)
        lt = len(topologies)
        if self.verbosity > 0:
            print('# n.topologies:',lt, 'tpldepth:', tpldepth)
        
        for tpl in topologies:
            enum_topologies.number_tree(tpl)
        
        scored = None
        if self.ordertopologies > 0:
            scored = [ (enum_topologies.diff_topologies(self.SKL_TOPO, tpl), i, tpl) for i, tpl in enumerate(topologies) ]
            scored.sort()
        else:
            scored = [ (1, i, tpl) for i, tpl in enumerate(topologies) ]
        
        if self.verbosity > 0:
            print('# Total topology generating and scoring time: {:.2f}'.format(time.time()-tgen))
        
        if lt > useassumps :
            oracle = self.satinit(cnf, True)
            print('# using assumption based solving')
        else:
            print('# no assumptions are used')
        
        satcalltimes = 0
        renewsatsolvertime = 0
        
        t = time.time()
        
        if self.verbosity > 0:
            print('# ', end = '')
        
        tobreak = False
        
        for i in range(len(scored)):
            if tobreak:
                break
            if self.verbosity > 0:
                print(';', end='')
            
            tpl = scored[i][2]
            assumps = self.settopology(tpl)
            
            if lt > useassumps:
                ti = time.time()
                res = oracle.solve(assumptions=assumps)
                if self.verbosity > 0:
                    tpi = time.time()-ti
                    satcalltimes += tpi
                    print('{:.2f}'.format(tpi), end='')
            else:
                ti = time.time()
                if oracle is not None:
                    oracle.delete()
                oracle = self.satinit(cnf, False)
                for aspt in assumps:
                    oracle.add_clause([aspt])
                tpi = time.time()-ti
                renewsatsolvertime += tpi
                
                ti = time.time()
                res = oracle.solve()
                if self.verbosity > 0:
                    tpi = time.time()-ti
                    satcalltimes += tpi
                    print('{:.2f}'.format(tpi), end='')
            
            if not res:
                if lt > useassumps:
                    core = oracle.get_core()
                    if core == None:
                        tobreak = True
                    elif len(core) < 0:
                        tobreak = True
                    else:
                        incore = False
                        for asp in assumps:
                            if asp in core:
                                incore = True
                                break
                        if not incore:
                            tobreak = True
            else:
                if self.verbosity > 0:
                    print('t{}'.format(i), end='')
                    #print('S', end='')
                tobreak = True
        if self.verbosity > 0:
            print('')
                
        if self.verbosity > 0:
            print('# Total topology test time: {:.2f}'.format(time.time()-t))
            print('# Total SAT Solving time {:.2f}'.format(satcalltimes))
            print('# Total Renew SAT Solver time {:.2f}'.format(renewsatsolvertime))
        
        return (oracle, res)
    #
    #--------------------------------------------------------------------------
    def solve_et_lssu(self):
        ubinit = self.N
        
        while self.N > 1 :
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
                
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                self.N = self.N-2
            
            else:
                cnf = self.encode()
                
                (oracle, res) = self.et_satsolve(cnf)
                    
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    oracle.delete()
                    return self.report_solution()
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    self.N = self.N-2
                    
                oracle.delete()
            
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def solve_et_lsus(self):
        stop = False
        ubinit = self.N
        self.N = 3
        while not stop :
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
                
            if self.N == ubinit:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                stop = True
            else:
                cnf = self.encode()
                    
                (oracle, res) = self.et_satsolve(cnf)
                        
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    self.N = self.N+2
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    stop = True
                    
                oracle.delete()
            
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def solve_et_bs(self):
        stop = False
        ubinit = self.N
        ub = self.N+2
        lb = 1
        while not stop :
            self.N = int((ub + lb)/2)
            if self.N % 2 == 0:
                self.N = self.N-1
            if self.verbosity > 0:
                print('# Testing #nodes =',self.N)
            
            if self.N == ubinit:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
            else:
                cnf = self.encode()
                
                (oracle, res) = self.et_satsolve(cnf)
                
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
                
                oracle.delete()
                
            if ub == lb + 2:
                stop = True
        
        return self.report_solution()
    #
    #--------------------------------------------------------------------------
    def solve(self):
        if self.algorithm == "lssu":
            return self.solve_lssu()
        elif self.algorithm == "lsus":
            return self.solve_lsus()
        elif self.algorithm == "bs":
            return self.solve_bs()
        elif self.algorithm == "prgsu":
            return self.solve_prgsu()
        elif self.algorithm == "prgus":
            return self.solve_prgus()
        elif self.algorithm == "et-lssu":
            return self.solve_et_lssu()
        elif self.algorithm == "et-lsus":
            return self.solve_et_lsus()
        elif self.algorithm == "et-bs":
            return self.solve_et_bs()
        else:
            print("Error:Unknown algorithm")
            return -1
