#!/usr/bin/env python3
#-*- coding:utf-8 -*-
##
## main.py
##
##  Created on: Jan 30, 2020
##      Author: Antonio Morgado
##      E-mail: ajrmorgado@gmail.com
##
# Copyright (C) 2020, Antonio Morgado, Mikolas Janota
#
# This file is part of dtfinder.
#
# dtfinder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dtfinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dtfinder.  If not, see <https://www.gnu.org/licenses/>.
#
#==============================================================================
#
# required imports


def stsCompare(sts):
    return sts['count']


class TrieNode(object):
    def __init__(self, nid):
        self.feat = None
        self.child0 = None
        self.child1 = None
        self.c = None
        self.nid = nid
        self.pid = None
    
    def print(self, nprefix = 0):
        prefix = ''
        for i in range(nprefix):
            prefix = '| ' + prefix
        print(prefix + '{}'.format(self.nid), end = '')
        if self.pid is None:
            print()
        else:
            print(' from', self.pid)
        if self.feat is not None:
            print(prefix + 'f:', self.feat)
        if self.child0 is not None:
            print(prefix + 'd0')
            self.child0.print(nprefix + 1)
        if self.child1 is not None:
            print(prefix + 'd1')
            self.child1.print(nprefix + 1)
        if self.c is not None:
            print(prefix + 'c:', self.c)
        
class Trie(object):
    def __init__(self):
        self.nid = 1
        self.root = TrieNode(0)
        self.featOrder = []
        self.nfeats = 0

    def setsamplesdin(self, samples):
        self.nfeats = len(samples[0])-1
        
        node = self.root
        feats = [i for i in range(self.nfeats)]
        idxsamples = [i for i in range(len(samples))]
        
        stack = [(node, feats, idxsamples)]
        
        while len(stack) > 0:
            (node, fs, ids) = stack.pop()
            #print('fs',fs, 'ids', ids)
            
            if len(fs) == 0:
                if len(ids) != 1:
                    print('Error: Empty node wihtout unique sample?')
                    #self.root.print()
                    exit(-1)
                node.c = samples[ids[0]][self.nfeats]
                
            else:
            
                assgn0f = [0 for x in range(len(fs))]
                assgn1f = [0 for x in range(len(fs))]
                for id in ids:
                    sample = samples[id]
                    for j in range(len(fs)):
                        if sample[fs[j]] == 0:
                            assgn0f[j] = assgn0f[j] + 1
                        else:
                            assgn1f[j] = assgn1f[j] + 1
                 
                maxf = fs[0]
                maxfcount = abs(assgn0f[0]-assgn1f[0])
                for j in range(1, len(fs)):
                    nf = fs[j]
                    nfcount = abs(assgn0f[j]-assgn1f[j])
                    if maxfcount < nfcount:
                        maxf = nf
                        maxfcount = nfcount
                 
                node.feat = maxf
                 
                nfs = [f for f in fs if f != maxf]
               
                ids0 = [id for id in ids if samples[id][maxf] == 0]
                if len(ids0) > 0:
                    node.child0 = TrieNode(self.nid)
                    self.nid = self.nid + 1
                    node.child0.pid = node.nid
                    stack.append((node.child0, nfs, ids0))
                    
                ids1 = [id for id in ids if samples[id][maxf] == 1]
                if len(ids1) > 0:
                    node.child1 = TrieNode(self.nid)
                    self.nid = self.nid + 1
                    node.child1.pid = node.nid
                    stack.append((node.child1, nfs, ids1))

    def mkstats(self, samples):
        self.nfeats = len(samples[0])-1
        assgn0f = [0 for x in range(self.nfeats)]
        assgn1f = [0 for x in range(self.nfeats)]
        for sample in samples:
            for i in range(self.nfeats):
                if sample[i] == 0:
                    assgn0f[i] = assgn0f[i] + 1
                else:
                    assgn1f[i] = assgn1f[i] + 1
        diff = [{'fo':i, 'count': abs(assgn0f[i]-assgn1f[i])} for i in range(self.nfeats)]
        diff.sort(reverse=True, key=stsCompare)
        self.featOrder = [diff[i]['fo'] for i in range(self.nfeats)]
        

    def setsamples(self, samples):
    
        self.mkstats(samples)
        for sample in samples:
            node = self.root
            for fo in self.featOrder:
                if node.feat is None:
                    node.feat = fo
                sval = sample[fo]
                if sval == 0:
                    if node.child0 is None:
                        node.child0 = TrieNode(self.nid)
                        self.nid = self.nid + 1
                        node.child0.pid = node.nid
                    node = node.child0
                else:
                    if node.child1 is None:
                        node.child1 = TrieNode(self.nid)
                        self.nid = self.nid + 1
                        node.child1.pid = node.nid
                    node = node.child1
            node.c = sample[self.nfeats]
                        
    def print(self):
        self.root.print()
        
if __name__ == "__main__":
    trie = Trie()
    samples = [[0,0,0,0],[0,0,1,0],[1,0,0,0]]
    print(samples)
    trie.setsamplesdin(samples)
    trie.print()

