#!/usr/bin/env python3
# File:  run_sklearn.py
# Author:  mikolas
# Created on:  Tue Dec 17 15:13:12 WET 2019
# Copyright (C) 2019, Mikolas Janota
# Copyright (C) 2020, Antonio Morgado, Mikolas Janota
#
# This file is part of dtfinder.
#
# dtfinder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dtfinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dtfinder.  If not, see <https://www.gnu.org/licenses/>.

import sys,subprocess,operator,time
import sklearn.tree
import getopt

def sklearn_tree_to_sol(clf):
    n_nodes = clf.tree_.node_count
    children_left = clf.tree_.children_left
    children_right = clf.tree_.children_right
    feature = clf.tree_.feature
    threshold = clf.tree_.threshold
    value = clf.tree_.value
    rv = []
    for node_id in range(n_nodes):
        if (children_left[node_id] != children_right[node_id]):
            rv.append("a {} {}".format(feature[node_id]+1, node_id+1))
            rv.append("l {} {}".format(node_id+1, children_left[node_id]+1))
            rv.append("r {} {}".format(node_id+1, children_right[node_id]+1))
        else:
            rv.append("c {} {}".format(node_id+1, 0 if value[node_id][0].item(0) else 1))
            # rv.append("# c {} {}".format(node_id+1, value[node_id]))
    return rv

def run_sk_learn(samples, seed=0):
    d = []
    v = []
    class_names = []
    class_set = set()
    for s in samples:
        d.append([ 1 if x==1 else -1 for x in s[:-1] ])
        v.append(1 if s[-1]==1 else -1)
    clf = sklearn.tree.DecisionTreeClassifier(random_state=seed)
    t0 = time.time()
    rv = clf.fit(d, v)
    upper_bound =  2*clf.get_n_leaves() - 1
    print ("# approx learn time: {:.2f}".format(time.time() - t0))
    return upper_bound, clf

def parse_samples(f):
    # trying to also parse Nina's stuff
    header = None
    samples = []
    for l in f:
        l = l.rstrip()
        if not l: continue
        if not header:
            split_sep = ',' if l.find(',') >= 0 else None
        s = l.split(split_sep)
        if header:
            assert(len(samples) == 0 or len(samples[0])== len(s))
            samples.append([int(l) for l in s])
        else:
            header = [int(l) for l in s] if len(s) == 1 else s
    return (header, samples)

if __name__ == "__main__":
    try:
        opts, args = getopt.getopt(sys.argv,"s",["seed="])
    except getopt.GetoptError:
        print('Error in options')
        sys.exit(1)
    seed = 0
    for opt, arg in opts:
        if opt in ("-s", "--seed"):
            seed = int(arg)
        else:
            print('Unknown option')
            sys.exit(1)

    header, samples = parse_samples(sys.stdin)
    node_count, skl = run_sk_learn(samples, seed=seed)
    print ("# number of nodes:", node_count)
    decision_tree = sklearn_tree_to_sol(skl)
    for l in decision_tree:
        print(l)
    exit(0)
